@@ -10,28 +10,51 @@ import re
 import stat
 import subprocess  # nosec: Needed for gitignore support.
 import sys
-from functools import lru_cache
+from dataclasses import dataclass, field
 from pathlib import Path
-from typing import Any, Callable, Dict, FrozenSet, Iterable, List, Optional, Pattern, Set, Tuple
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    FrozenSet,
+    Iterable,
+    List,
+    Optional,
+    Pattern,
+    Set,
+    Tuple,
+    Type,
+    Union,
+)
 from warnings import warn
 
-from . import stdlibs
-from ._future import dataclass, field
-import toml
+from . import sorting, stdlibs
 from .exceptions import (
     FormattingPluginDoesNotExist,
     InvalidSettingsPath,
     ProfileDoesNotExist,
+    SortingFunctionDoesNotExist,
     UnsupportedSettings,
 )
-from .profiles import profiles
+from .profiles import profiles as profiles
 from .sections import DEFAULT as SECTION_DEFAULTS
 from .sections import FIRSTPARTY, FUTURE, LOCALFOLDER, STDLIB, THIRDPARTY
+from .utils import Trie
 from .wrap_modes import WrapModes
 from .wrap_modes import from_string as wrap_mode_from_string
 
-_SHEBANG_RE = re.compile(br"^#!.*\bpython[23w]?\b")
-SUPPORTED_EXTENSIONS = frozenset({"py", "pyi", "pyx", "pxd"})
+if TYPE_CHECKING:
+    tomllib: Any
+else:
+    if sys.version_info >= (3, 11):
+        import tomllib
+    else:
+        import tomli as tomllib
+
+_SHEBANG_RE = re.compile(rb"^#!.*\bpython[23w]?\b")
+CYTHON_EXTENSIONS = frozenset({"pyx", "pxd"})
+SUPPORTED_EXTENSIONS = frozenset({"py", "pyi", *CYTHON_EXTENSIONS})
 BLOCKED_EXTENSIONS = frozenset({"pex"})
 FILE_SKIP_COMMENTS: Tuple[str, ...] = (
     "isort:" + "skip_file",
@@ -68,6 +91,8 @@ DEFAULT_SKIP: FrozenSet[str] = frozenset(
         ".pants.d",
         ".direnv",
         "node_modules",
+        "__pypackages__",
+        ".pytype",
     }
 )
 
@@ -81,6 +106,7 @@ CONFIG_SECTIONS: Dict[str, Tuple[str, ...]] = {
 FALLBACK_CONFIG_SECTIONS: Tuple[str, ...] = ("isort", "tool:isort", "tool.isort")
 
 IMPORT_HEADING_PREFIX = "import_heading_"
+IMPORT_FOOTER_PREFIX = "import_footer_"
 KNOWN_PREFIX = "known_"
 KNOWN_SECTION_MAPPING: Dict[str, str] = {
     STDLIB: "STANDARD_LIBRARY",
@@ -121,7 +147,9 @@ class _Config:
     py_version: str = "3"
     force_to_top: FrozenSet[str] = frozenset()
     skip: FrozenSet[str] = DEFAULT_SKIP
+    extend_skip: FrozenSet[str] = frozenset()
     skip_glob: FrozenSet[str] = frozenset()
+    extend_skip_glob: FrozenSet[str] = frozenset()
     skip_gitignore: bool = False
     line_length: int = 79
     wrap_length: int = 0
@@ -150,10 +178,12 @@ class _Config:
     single_line_exclusions: Tuple[str, ...] = ()
     default_section: str = THIRDPARTY
     import_headings: Dict[str, str] = field(default_factory=dict)
+    import_footers: Dict[str, str] = field(default_factory=dict)
     balanced_wrapping: bool = False
     use_parentheses: bool = False
     order_by_type: bool = True
     atomic: bool = False
+    lines_before_imports: int = -1
     lines_after_imports: int = -1
     lines_between_sections: int = 1
     lines_between_types: int = 0
@@ -205,14 +235,26 @@ class _Config:
     namespace_packages: FrozenSet[str] = frozenset()
     follow_links: bool = True
     indented_import_headings: bool = True
+    honor_case_in_force_sorted_sections: bool = False
+    sort_relative_in_force_sorted_sections: bool = False
+    overwrite_in_place: bool = False
+    reverse_sort: bool = False
+    star_first: bool = False
+    import_dependencies = Dict[str, str]
+    git_ls_files: Dict[Path, Set[str]] = field(default_factory=dict)
+    format_error: str = "{error}: {message}"
+    format_success: str = "{success}: {message}"
+    sort_order: str = "natural"
+    sort_reexports: bool = False
+    split_on_trailing_comma: bool = False
 
-    def __post_init__(self):
+    def __post_init__(self) -> None:
         py_version = self.py_version
         if py_version == "auto":  # pragma: no cover
             if sys.version_info.major == 2 and sys.version_info.minor <= 6:
                 py_version = "2"
             elif sys.version_info.major == 3 and (
-                sys.version_info.minor <= 5 or sys.version_info.minor >= 9
+                sys.version_info.minor <= 5 or sys.version_info.minor >= 12
             ):
                 py_version = "3"
             else:
@@ -233,6 +275,9 @@ class _Config:
                 self, "known_standard_library", frozenset(getattr(stdlibs, self.py_version).stdlib)
             )
 
+        if self.multi_line_output == WrapModes.VERTICAL_GRID_GROUPED_NO_COMMA:  # type: ignore
+            vertical_grid_grouped = WrapModes.VERTICAL_GRID_GROUPED  # type: ignore
+            object.__setattr__(self, "multi_line_output", vertical_grid_grouped)
         if self.force_alphabetical_sort:
             object.__setattr__(self, "force_alphabetical_sort_within_sections", True)
             object.__setattr__(self, "no_sections", True)
@@ -244,7 +289,7 @@ class _Config:
                 f"{self.wrap_length} > {self.line_length}."
             )
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return id(self)
 
 
@@ -257,10 +302,14 @@ class Config(_Config):
         settings_file: str = "",
         settings_path: str = "",
         config: Optional[_Config] = None,
-        **config_overrides,
+        **config_overrides: Any,
     ):
         self._known_patterns: Optional[List[Tuple[Pattern[str], str]]] = None
         self._section_comments: Optional[Tuple[str, ...]] = None
+        self._section_comments_end: Optional[Tuple[str, ...]] = None
+        self._skips: Optional[FrozenSet[str]] = None
+        self._skip_globs: Optional[FrozenSet[str]] = None
+        self._sorting_function: Optional[Callable[..., List[str]]] = None
 
         if config:
             config_vars = vars(config).copy()
@@ -268,7 +317,11 @@ class Config(_Config):
             config_vars["py_version"] = config_vars["py_version"].replace("py", "")
             config_vars.pop("_known_patterns")
             config_vars.pop("_section_comments")
-            super().__init__(**config_vars)  # type: ignore
+            config_vars.pop("_section_comments_end")
+            config_vars.pop("_skips")
+            config_vars.pop("_skip_globs")
+            config_vars.pop("_sorting_function")
+            super().__init__(**config_vars)
             return
 
         # We can't use self.quiet to conditionally show warnings before super.__init__() is called
@@ -339,6 +392,7 @@ class Config(_Config):
 
         known_other = {}
         import_headings = {}
+        import_footers = {}
         for key, value in tuple(combined_config.items()):
             # Collect all known sections beyond those that have direct entries
             if key.startswith(KNOWN_PREFIX) and key not in (
@@ -375,6 +429,8 @@ class Config(_Config):
                         )
             if key.startswith(IMPORT_HEADING_PREFIX):
                 import_headings[key[len(IMPORT_HEADING_PREFIX) :].lower()] = str(value)
+            if key.startswith(IMPORT_FOOTER_PREFIX):
+                import_footers[key[len(IMPORT_FOOTER_PREFIX) :].lower()] = str(value)
 
             # Coerce all provided config values into their correct type
             default_value = _DEFAULT_SETTINGS.get(key, None)
@@ -409,9 +465,12 @@ class Config(_Config):
         else:
             src_paths: List[Path] = []
             for src_path in combined_config.get("src_paths", ()):
-                full_path = path_root / src_path
-                if full_path not in src_paths:
-                    src_paths.append(full_path)
+                full_paths = (
+                    path_root.glob(src_path) if "*" in str(src_path) else [path_root / src_path]
+                )
+                for path in full_paths:
+                    if path not in src_paths:
+                        src_paths.append(path)
 
             combined_config["src_paths"] = tuple(src_paths)
 
@@ -441,7 +500,8 @@ class Config(_Config):
                 warn(
                     "W0503: Deprecated config options were used: "
                     f"{', '.join(deprecated_options_used)}."
-                    "Please see the 5.0.0 upgrade guide: bit.ly/isortv5."
+                    "Please see the 5.0.0 upgrade guide: "
+                    "https://pycqa.github.io/isort/docs/upgrade_guides/5.0.0.html"
                 )
 
         if known_other:
@@ -450,6 +510,10 @@ class Config(_Config):
             for import_heading_key in import_headings:
                 combined_config.pop(f"{IMPORT_HEADING_PREFIX}{import_heading_key}")
             combined_config["import_headings"] = import_headings
+        if import_footers:
+            for import_footer_key in import_footers:
+                combined_config.pop(f"{IMPORT_FOOTER_PREFIX}{import_footer_key}")
+            combined_config["import_footers"] = import_footers
 
         unsupported_config_errors = {}
         for option in set(combined_config.keys()).difference(
@@ -464,9 +528,9 @@ class Config(_Config):
         if unsupported_config_errors:
             raise UnsupportedSettings(unsupported_config_errors)
 
-        super().__init__(sources=tuple(sources), **combined_config)  # type: ignore
+        super().__init__(sources=tuple(sources), **combined_config)
 
-    def is_supported_filetype(self, file_name: str):
+    def is_supported_filetype(self, file_name: str) -> bool:
         _root, ext = os.path.splitext(file_name)
         ext = ext.lstrip(".")
         if ext in self.supported_extensions:
@@ -492,6 +556,43 @@ class Config(_Config):
         else:
             return bool(_SHEBANG_RE.match(line))
 
+    def _check_folder_git_ls_files(self, folder: str) -> Optional[Path]:
+        env = {**os.environ, "LANG": "C.UTF-8"}
+        try:
+            topfolder_result = subprocess.check_output(  # nosec # skipcq: PYL-W1510
+                ["git", "-C", folder, "rev-parse", "--show-toplevel"], encoding="utf-8", env=env
+            )
+        except subprocess.CalledProcessError:
+            return None
+
+        git_folder = Path(topfolder_result.rstrip()).resolve()
+
+        # files committed to git
+        tracked_files = (
+            subprocess.check_output(  # nosec # skipcq: PYL-W1510
+                ["git", "-C", str(git_folder), "ls-files", "-z"],
+                encoding="utf-8",
+                env=env,
+            )
+            .rstrip("\0")
+            .split("\0")
+        )
+        # files that haven't been committed yet, but aren't ignored
+        tracked_files_others = (
+            subprocess.check_output(  # nosec # skipcq: PYL-W1510
+                ["git", "-C", str(git_folder), "ls-files", "-z", "--others", "--exclude-standard"],
+                encoding="utf-8",
+                env=env,
+            )
+            .rstrip("\0")
+            .split("\0")
+        )
+
+        self.git_ls_files[git_folder] = {
+            str(git_folder / Path(f)) for f in tracked_files + tracked_files_others
+        }
+        return git_folder
+
     def is_skipped(self, file_path: Path) -> bool:
         """Returns True if the file and/or folder should be skipped based on current settings."""
         if self.directory and Path(self.directory) in file_path.resolve().parents:
@@ -501,21 +602,11 @@ class Config(_Config):
 
         os_path = str(file_path)
 
-        if self.skip_gitignore:
-            if file_path.name == ".git":  # pragma: no cover
-                return True
-
-            result = subprocess.run(  # nosec # skipcq: PYL-W1510
-                ["git", "-C", str(file_path.parent), "check-ignore", "--quiet", os_path]
-            )
-            if result.returncode == 0:
-                return True
-
         normalized_path = os_path.replace("\\", "/")
         if normalized_path[1:2] == ":":
             normalized_path = normalized_path[2:]
 
-        for skip_path in self.skip:
+        for skip_path in self.skips:
             if posixpath.abspath(normalized_path) == posixpath.abspath(
                 skip_path.replace("\\", "/")
             ):
@@ -523,21 +614,44 @@ class Config(_Config):
 
         position = os.path.split(file_name)
         while position[1]:
-            if position[1] in self.skip:
+            if position[1] in self.skips:
                 return True
             position = os.path.split(position[0])
 
-        for glob in self.skip_glob:
-            if fnmatch.fnmatch(file_name, glob) or fnmatch.fnmatch("/" + file_name, glob):
+        for sglob in self.skip_globs:
+            if fnmatch.fnmatch(file_name, sglob) or fnmatch.fnmatch("/" + file_name, sglob):
                 return True
 
         if not (os.path.isfile(os_path) or os.path.isdir(os_path) or os.path.islink(os_path)):
             return True
 
+        if self.skip_gitignore:
+            if file_path.name == ".git":  # pragma: no cover
+                return True
+
+            git_folder = None
+
+            file_paths = [file_path, file_path.resolve()]
+            for folder in self.git_ls_files:
+                if any(folder in path.parents for path in file_paths):
+                    git_folder = folder
+                    break
+            else:
+                git_folder = self._check_folder_git_ls_files(str(file_path.parent))
+
+            # git_ls_files are good files you should parse. If you're not in the allow list, skip.
+
+            if (
+                git_folder
+                and not file_path.is_dir()
+                and str(file_path.resolve()) not in self.git_ls_files[git_folder]
+            ):
+                return True
+
         return False
 
     @property
-    def known_patterns(self):
+    def known_patterns(self) -> List[Tuple[Pattern[str], str]]:
         if self._known_patterns is not None:
             return self._known_patterns
 
@@ -568,6 +682,53 @@ class Config(_Config):
         self._section_comments = tuple(f"# {heading}" for heading in self.import_headings.values())
         return self._section_comments
 
+    @property
+    def section_comments_end(self) -> Tuple[str, ...]:
+        if self._section_comments_end is not None:
+            return self._section_comments_end
+
+        self._section_comments_end = tuple(f"# {footer}" for footer in self.import_footers.values())
+        return self._section_comments_end
+
+    @property
+    def skips(self) -> FrozenSet[str]:
+        if self._skips is not None:
+            return self._skips
+
+        self._skips = self.skip.union(self.extend_skip)
+        return self._skips
+
+    @property
+    def skip_globs(self) -> FrozenSet[str]:
+        if self._skip_globs is not None:
+            return self._skip_globs
+
+        self._skip_globs = self.skip_glob.union(self.extend_skip_glob)
+        return self._skip_globs
+
+    @property
+    def sorting_function(self) -> Callable[..., List[str]]:
+        if self._sorting_function is not None:
+            return self._sorting_function
+
+        if self.sort_order == "natural":
+            self._sorting_function = sorting.naturally
+        elif self.sort_order == "native":
+            self._sorting_function = sorted
+        else:
+            available_sort_orders = ["natural", "native"]
+            import pkg_resources
+
+            for sort_plugin in pkg_resources.iter_entry_points("isort.sort_function"):
+                available_sort_orders.append(sort_plugin.name)
+                if sort_plugin.name == self.sort_order:
+                    self._sorting_function = sort_plugin.load()
+                    break
+            else:
+                raise SortingFunctionDoesNotExist(self.sort_order, available_sort_orders)
+
+        return self._sorting_function
+
     def _parse_known_pattern(self, pattern: str) -> List[str]:
         """Expand pattern if identified as a directory and return found sub packages"""
         if pattern.endswith(os.path.sep):
@@ -582,8 +743,10 @@ class Config(_Config):
         return patterns
 
 
-def _get_str_to_type_converter(setting_name: str) -> Callable[[str], Any]:
-    type_converter: Callable[[str], Any] = type(_DEFAULT_SETTINGS.get(setting_name, ""))
+def _get_str_to_type_converter(setting_name: str) -> Union[Callable[[str], Any], Type[Any]]:
+    type_converter: Union[Callable[[str], Any], Type[Any]] = type(
+        _DEFAULT_SETTINGS.get(setting_name, "")
+    )
     if type_converter == WrapModes:
         type_converter = wrap_mode_from_string
     return type_converter
@@ -606,7 +769,6 @@ def _abspaths(cwd: str, values: Iterable[str]) -> Set[str]:
     return paths
 
 
-@lru_cache()
 def _find_config(path: str) -> Tuple[str, Dict[str, Any]]:
     current_directory = path
     tries = 0
@@ -639,19 +801,47 @@ def _find_config(path: str) -> Tuple[str, Dict[str, Any]]:
     return (path, {})
 
 
-@lru_cache()
-def _get_config_data(file_path: str, sections: Tuple[str]) -> Dict[str, Any]:
+def find_all_configs(path: str) -> Trie:
+    """
+    Looks for config files in the path provided and in all of its sub-directories.
+    Parses and stores any config file encountered in a trie and returns the root of
+    the trie
+    """
+    trie_root = Trie("default", {})
+
+    for dirpath, _, _ in os.walk(path):
+        for config_file_name in CONFIG_SOURCES:
+            potential_config_file = os.path.join(dirpath, config_file_name)
+            if os.path.isfile(potential_config_file):
+                config_data: Dict[str, Any]
+                try:
+                    config_data = _get_config_data(
+                        potential_config_file, CONFIG_SECTIONS[config_file_name]
+                    )
+                except Exception:
+                    warn(f"Failed to pull configuration information from {potential_config_file}")
+                    config_data = {}
+
+                if config_data:
+                    trie_root.insert(potential_config_file, config_data)
+                    break
+
+    return trie_root
+
+
+def _get_config_data(file_path: str, sections: Tuple[str, ...]) -> Dict[str, Any]:
     settings: Dict[str, Any] = {}
 
-    with open(file_path) as config_file:
-        if file_path.endswith(".toml"):
-            config = toml.load(config_file)
-            for section in sections:
-                config_section = config
-                for key in section.split("."):
-                    config_section = config_section.get(key, {})
-                settings.update(config_section)
-        else:
+    if file_path.endswith(".toml"):
+        with open(file_path, "rb") as bin_config_file:
+            config = tomllib.load(bin_config_file)
+        for section in sections:
+            config_section = config
+            for key in section.split("."):
+                config_section = config_section.get(key, {})
+            settings.update(config_section)
+    else:
+        with open(file_path, encoding="utf-8") as config_file:
             if file_path.endswith(".editorconfig"):
                 line = "\n"
                 last_position = config_file.tell()
@@ -664,18 +854,22 @@ def _get_config_data(file_path: str, sections: Tuple[str]) -> Dict[str, Any]:
 
             config = configparser.ConfigParser(strict=False)
             config.read_file(config_file)
-            for section in sections:
-                if section.startswith("*.{") and section.endswith("}"):
-                    extension = section[len("*.{") : -1]
-                    for config_key in config.keys():
-                        if config_key.startswith("*.{") and config_key.endswith("}"):
-                            if extension in map(
-                                lambda text: text.strip(), config_key[len("*.{") : -1].split(",")
-                            ):
-                                settings.update(config.items(config_key))
+        for section in sections:
+            if section.startswith("*.{") and section.endswith("}"):
+                extension = section[len("*.{") : -1]
+                for config_key in config.keys():
+                    if (
+                        config_key.startswith("*.{")
+                        and config_key.endswith("}")
+                        and extension
+                        in map(
+                            lambda text: text.strip(), config_key[len("*.{") : -1].split(",")  # type: ignore # noqa
+                        )
+                    ):
+                        settings.update(config.items(config_key))
 
-                elif config.has_section(section):
-                    settings.update(config.items(section))
+            elif config.has_section(section):
+                settings.update(config.items(section))
 
     if settings:
         settings["source"] = file_path
